program Hess2d
 !!*** Computes Gauss-Newton Hessian in 2D using point-spread functions
  use sep
  use declare_variables_mod
  use read_mod, only : read_
  use write_mod, only : write_
  use src_rcv_interp_mod
  use vshot_mod, only : vshot_foo
  use rshot_mod, only : rshot_foo
  use fwd_propagation_mod, only : src_prop_fwd
  use bwd_propagation_mod, only : src_rcv_prop_bwd
  use rtm_mod, only : rtm_image
  use born_mod, only : born_foo
!  use H_mod
  use lint2_mod
  use taper_mod
  use skull_mod
  use assertions
  implicit none
  integer                :: j, jz, jx, jzc, jxc, jlagz, jlagx
  integer                :: nzc, nxc, nlagz, nlagx, nfilt, nreg, incPSFx, incPSFz
  integer                :: time(8), seed(2), allocation_status, stat
  real                   :: eps, x, z, o1, o2, d1, d2
  real, allocatable      :: data_(:)
  real, allocatable      :: model(:,:)
  real, allocatable      :: Hgn(:,:,:,:)
!  real, pointer          :: xz(:,:)
  real, allocatable, dimension(:,:) :: psf_approx, m, d
  type(src_prop_fwd)     :: fwd   ! Arrays and procedures for source forward propagation
  type(src_rcv_prop_bwd) :: bwd   ! Arrays and procedures for source/receiver backward propagation
  type(born_foo)         :: born  ! Arrays and procedures for imaging conditions
  type(rtm_image)        :: image ! Arrays and procedures for imaging conditions
  type(read_)            :: par   ! Contains the input parameters and reading procedures
  type(write_)           :: outp  ! Contains the ouput parameters and reading procedures
  type(vshot_foo)        :: vel
  type(rshot_foo)        :: ref

  !! Initialize sep libraries
  call sep_init()
  call init_3d()
  
  !! Initialize timing
  call date_and_time(values=timer0)
  
  !! Read the input history file
  call par%read_parameters()

  !! Read lags for off-diagonal elements of the GN Hessian
  call from_param('nlagz',nlagz,0)
  call from_param('nlagx',nlagx,0)

  !! Read increments for PSF seeding
  call from_param('incPSFz',incPSFz,20)
  call from_param('incPSFx',incPSFx,20)

  !! De-activate energy imaging conditions
  par%eic=F
  
  !! Set apertures to zero
  par%apertureL=0.
  par%apertureR=0.
  
  !! Read data history
  call par%read_history()
  
  !! Read data header
  call par%read_headers()
  
  !! Read velocity
  call par%read_velocity()

  !! Read reflectivity
  call par%read_reflectivity()

  !! Seed the PSFs (overwrite reflectivity)
  par%refl=0.
  nxc=0
  associate(x1=>par%first_cdp,x2=>par%last_cdp)
    do jx=x1+incPSFx,x2,incPSFx ! I found border effects if I put
                                ! a spike at lateral boundaries
      nxc=nxc+1
      nzc=0
      do jz=1+incPSFz,par%nz,incPSFz
        nzc=nzc+1
        par%refl(jz,jx) =1.
      end do
    end do
    nzc=nzc+1
    nxc=nxc+1
    allocate(model(nzc,nxc),             &
             data_(par%nz*par%nx), &
             Hgn(par%nz,x1:x2,-nlagz:nlagz,-nlagx:nlagx),  &
!             xz(par%nz*par%nx,2),  &
             stat=allocation_status)
    if(allocation_status /= 0) then
      write(0,*) 'Hess2d: Cannot allocate arrays for Hessian!'
      call skull()
    end if
    model=0.
    data_=0.
    Hgn=0.
    allocate(m(par%nz,par%nx),&
             d(par%nz,par%nx),&
             psf_approx(par%nz,x1:x2))
  end associate
 
  !! Determine time sampling according to Courant-Fredrich-Lewis condition
  call par%courant_stuff()
  
  !! Obtain data geometry and apertures from input parameters
  call par%data_geometry()

  !! Read and verbose source
  call par%read_src()
  
  !! Verbose some useful parameters
  call par%verbose_par()
  write(0,*) 'nlagz: ',nlagz 
  write(0,*) 'nlagx: ',nlagx 
  write(0,*) 'incPSFz: ',incPSFz
  write(0,*) 'incPSFx: ',incPSFx
 
  !! Initialize parameters for forward and backward propagations
  call fwd%fwd_prop_init(par)
  call bwd%bwd_prop_init(par)
  
  !! Initialize first derivatives parameters
  !call fwd%grad_init()
  
  !! Initialize variables for shot velocities
  call vel%vshot_init(par) 
  
  !! Initialize variables for reflectivity
  call ref%rshot_init(par) 
  
  !! Initialize tapering for receiver propagation
  call taper_init(par,use_taper)
  
  !! Initialize derivatives coeffiecients
  call fwd%lap_init()
  
  !! Initialize parameters for born modeling
  call born%born_init(par,fwd,sx,gx)

  !! Initialize parameters for imaging conditions
  call image%rtm_init(par,fwd,bwd,sx,gx)
 
  !! Initialize some variables for the main loop
  par%first_geof = 1
  par%ntSnap=0
  
  !! Initialize writing module array
  call outp%write_init(par) 

  !! This is the main loop for Hessian computation
  write(0,*) ''
  write(0,*) '********************** MAIN LOOP BEGINS! ***********************'
  write(0,*) ''
  loop_over_sources: do jsources = par%first_source,par%last_source
    !! Get number of geophone traces in shot gather (in some datasets is variable)
    par%ngeof = source_index(jsources,1)
    par%last_geof = par%first_geof+par%ngeof-1
  
    !! Create velocity frames for source wavefield propagation
    call vel%vshot(par,fwd%Vshot,par%sgx(:,jsources), .true.)
    call vel%vshot(par,fwd%Vshotr,par%sgx(:,jsources),.false.)
    call sep_write(fwd%Vshot,'vrandom')
    if(.not. allocated(bwd%Vshot))  allocate(bwd%Vshot, source=fwd%Vshot)
    if(.not. allocated(bwd%Vshotr)) allocate(bwd%Vshotr,source=fwd%Vshotr)

    !! Compute shot position and report progress
    fwd%shot_pos = 1+floor((sx(par%trace_shot)-gx(par%trace_shot))/par%dx)
    bwd%shot_pos = 1+floor((sx(par%trace_shot)-gx(par%trace_shot))/par%dx)
    write(0,'(a,i7,i7,i7,i7,i7,i7,i5)')     &
    'No. sources,curr source,shot pos.,source_index,first_geof,last_geof,ngeof', &
    par%nsources,jsources,fwd%shot_pos,source_index(jsources,2),&
    par%first_geof,par%last_geof,par%ngeof

    !! Perform Born Modeling for each shot gather
    call born%born(par,fwd,vel,ref,jsources)

    !! Pass the Born modeled PSF to input for RTM
    par%rcv=born%dsynt

    !! Perform Reverse-Time Migration in each shot gather
    call image%rtm(par,fwd,bwd,jsources)

    !! Construct the images
    associate(x1=>1-par%napL,x2=>par%ngx+par%napR)
      call image%build_images(par,par%sgx(x1:x2,jsources),F)
    end associate

    !! Advance to the next receiver and the corresponding traces
    par%trace_shot = par%trace_shot+par%ngeof
    par%first_geof = par%last_geof+1
  end do loop_over_sources
  write(0,*) ''
  write(0,*) '********************** MAIN LOOP FINISHES! ***********************'
  write(0,*) ''

!      associate(x1=>par%first_cdp,x2=>par%last_cdp)
!        do jx=x1+incPSFx,x2,incPSFx
!          do jz=1+incPSFz,par%nz,incPSFz
!            image%ICorr(jz,jx)=image%ICorr(jz,jx)-99999999.
!          end do
!        end do
!      end associate
  !! Write the point-spread functions
  write(0,*) 'Write point spread functions'
  call outp%write_history(par,image,.false.)

!  associate(x1=>par%first_cdp,x2=>par%last_cdp)
!    do jx=x1,x2
!      x=(jx-1)*par%dx
!      do jz=1,par%nz
!        z=(jz-1)*par%dz
!        j=(jx-x1)*par%nz+jz
!        xz(j,1)=z
!        xz(j,2)=x
!      end do
!    end do
!    d1=incPSFz*par%dz
!    d2=incPSFx*par%dx
!    o1=incPSFz*par%dz
!    o2=(x1+incPSFx-1)*par%dx
!  end associate
 
!  do jlagx = -nlagx,nlagx
!    do jlagz = -nlagz,nlagz
!
!      !! Extrac elements of the GN Hessian
!      associate(x1=>par%first_cdp,x2=>par%last_cdp)
!        jxc=0
!        model=0.
!        do jx=x1+incPSFx,x2,incPSFx
!          jxc=jxc+1
!          jzc=0
!          do jz=1+incPSFz,par%nz,incPSFz
!            jzc=jzc+1
!            model(jzc,jxc)=image%Icorr(jz+jlagz,jx+jlagx)
!!write(0,*) 'Icorr,jzc,jxc,jz,jx',image%Icorr(jz+jlagz,jx+jlagx),jzc,jxc,jz,jx
!!write(0,*) 'model,jzc,jxc,jz,jx',model(jzc,jxc),jzc,jxc,jz,jx
!          end do
!        end do
!      end associate
!      !! Interpolate the diagonal of the Hessian
!      data_=0.
!      call lint2_init(nzc,nxc,o1,d1,o2,d2,xz)
!      call lint2_lop2(F,T,model,data_)
!      Hgn(:,:,jlagz,jlagx)=reshape(data_,(/par%nz,par%nx/))
!    end do
!  end do

!  !! Write the interpolated Hessian
!  associate(nz=>par%nz,x1=>par%first_cdp,x2=>par%last_cdp)
!    write(0,*) 'Write the interpolated Hessian'
!    call outp%write_history(par,image,Hgn,nz,x1,x2,nlagz,nlagx)
!  end associate
!
!  !! TEST: Check the performance of the Hgn operator to reproduce the PSF
!    m=par%refl
!    call H_init(par,F,Hgn)
!    call H_op2(F,F,m,d) 
!    image%ICorr=d
!    write(0,*) 'Write the approximated point spread functions'
!    call outp%write_history(par,image,F,"psf_approx")

  !! Get the timing
  call date_and_time(values=timer1)
  timer0 = timer1-timer0
  timer_sum = timer_sum + (((timer0(3)*24+timer0(5))*60+timer0(6))*60+timer0(7))*1000+timer0(8)
  call to_history("timer sum (sec)", timer_sum/1000)
  call to_history("timer sum (min)", timer_sum/1000/60)
  write(0,*) "timer sum (sec)", timer_sum/1000
  write(0,*) "timer sum (min)", timer_sum/1000/60
  
  !! Finale!
  write(0,*) ' '
  write(0,*) 'Finishing the Hess2d program normally!'

end program
