program Linear_test
 !!*** Perform linearity tests
  use sep
  use declare_variables_mod
  use read_mod, only : read_
  use write_mod, only : write_
  use src_rcv_interp_mod
  use vshot_mod, only : vshot_foo
  use rshot_mod, only : rshot_foo
  use fwd_propagation_mod, only : src_prop_fwd
  use bwd_propagation_mod, only : src_rcv_prop_bwd
  use nonlin_mod, only : nonlin_foo
  use rtm_mod, only : rtm_image
  use born_mod, only : born_foo
  use wemva_mod, only : wemva_foo
  use taper_mod
  use skull_mod
  use assertions
  use born_op_mod
  use wemva_op_mod
  implicit none
  integer                 :: time(8), seed(2)
  integer :: jt, stat
  real*8 :: dpData, dpRefl, err_abs, err_rel
  real :: S2cte
  real, allocatable :: rand_var(:,:)
  real, dimension(:), allocatable :: d1, d2, diff, src_tmp
  real, dimension(:,:), allocatable :: dI, dV, dS2, dexact, dapprox, temp1, refl_temp
  real, dimension(:,:,:), allocatable :: dpxz
  character :: op
  character(len=100) :: wavelet_nl_fn
  type(src_prop_fwd)     :: fwd    ! Arrays and procedures for source forward propagation
  type(src_rcv_prop_bwd) :: bwd    ! Arrays and procedures for source/receiver backward propagation
  type(nonlin_foo)       :: nonlin ! Arrays and procedures for non-linear modeling
  type(born_foo)         :: born   ! Arrays and procedures for imaging conditions
  type(rtm_image)        :: image  ! Arrays and procedures for imaging conditions
  type(wemva_foo)        :: wemva  ! Arrays and procedures for imaging conditions
  type(read_)            :: par    ! Contains the input parameters and reading procedures
  type(write_)           :: outp   ! Contains the ouput parameters and reading procedures
  type(vshot_foo)        :: vel
  type(rshot_foo)        :: ref

  !! Initialize sep libraries
  call sep_init()
  call init_3d()
  
  !! Initialize timing
  call date_and_time(values=timer0)
  
  !! Read the input history file
  call par%read_parameters()

  !! Additional parameters
  call from_param('op',op,'B')
  
!! De-activate energy imaging conditions for lntest
  par%eic=F
  
  !! Overriden parameters for do-product test
  par%one_src='y'
  
  !! Read data history
  call par%read_history()
  
  !! Read data header
  call par%read_headers()
  
  !! Read velocity
  call par%read_velocity()
  par%V_int=2500.
  par%Vcte = sum(par%V_int(:,:))/(par%nx*par%nz)
  par%Vmin = minval(par%V_int)
  par%Vmax = maxval(par%V_int)

  !! Read reflectivity
  call par%read_reflectivity()

  !! Prepare velocity perturbed fields V+dV
  allocate(dV, mold=par%V_int)
  allocate(dS2,mold=par%refl)

  dV=0.
  dV(par%nz/2,par%first_cdp+par%nx/2)=par%Vcte*0.03
  write(0,*) 'minval(dV)',minval(dV)
  write(0,*) 'maxval(dV)',maxval(dV)

  dS2=0.
  dS2=(1./(par%V_int+dV)**2-1./par%V_int**2)
  par%refl=dS2
  write(0,*) 'minval(dS2)',minval(dS2)
  write(0,*) 'maxval(dS2)',maxval(dS2)

  par%V_int=par%V_int+dV
  !! Determine time sampling according to Courant-Fredrich-Lewis condition
  call par%courant_stuff()
  par%V_int=par%V_int-dV

  !! Obtain data geometry and apertures from input parameters
  call par%data_geometry()
  par%nsources=1
  
  !! Read and verbose source
  select case(op)
    case('W')
      !! Read the 2nd. derivative of the source for RM and WEMVA
      call par%read_src()
    case('B')
      !! Read the original source for non-linear modeling
      call from_param('wavelet_nl_fn',wavelet_nl_fn)
      call par%read_src(T,wavelet_nl_fn)
      allocate(src_tmp,source=par%src)
  
      !! Read the 2nd. derivative of the source for Born modeling 
      call par%read_src()
    case default
      call skull()
      error stop "Linear_test(1): Unknown operator" 
  end select

  !! Verbose some useful parameters
  call par%verbose_par()
  
  !! Initialize parameters for forward and backward propagations
  call fwd%fwd_prop_init(par)
  call bwd%bwd_prop_init(par)
  
  !! Initialize variables for shot velocities
  call vel%vshot_init(par) 
  
  !! Initialize variables for reflectivity
  call ref%rshot_init(par) 
  
  !! Initialize tapering for receiver propagation
  call taper_init(par,use_taper)
  
  !! Initialize derivatives coeffiecients
  call fwd%lap_init()
 
  select case(op) 
    case('W') 
      !! Initialize parameters for rtm operator
      call image%rtm_init(par,fwd,bwd,sx,gx)
  
      !! Initialize parameters for wemva
      call wemva%wemva_init(par,fwd,sx,gx)
  
      !! Prepare arrays for the output
      allocate(dexact ,mold=image%ICorr)
      allocate(dapprox,mold=wemva%IPert)
    case('B')
      !! Initialize parameters for non-linear modeling
      call nonlin%nonlin_init(par,fwd,sx,gx)
  
      !! Initialize parameters for forward operator
      call born%born_init(par,fwd,sx,gx)
  
      !! Prepare arrays for the output
      allocate(dexact ,mold=born%dsynt)
      allocate(dapprox,mold=born%dsynt)
    case default
      call skull()
      error stop "Linear_test(2): Unknown operator" 
  end select
 
  !! Initialize some variables for the main loop
  par%first_geof = 1
  par%ntSnap=0
  
  !! Initialize writing module array
  call outp%write_init(par) 

  !! This is the main loop to migrate the shots
  write(0,*) ''
  write(0,*) '********************** MAIN LOOP BEGINS! ***********************'
  write(0,*) ''
  loop_over_sources: do jsources = par%first_source,par%last_source
    !! Get number of geophone traces in shot gather (in some datasets is variable)
    par%ngeof = source_index(jsources,1)
    par%last_geof = par%first_geof+par%ngeof-1
  
    !! Create velocity frames for source wavefield propagation
    call vel%vshot(par,fwd%Vshot, par%sgx(:,jsources),T)
    fwd%Vshotr=fwd%Vshot
!    call vel%vshot(par,fwd%Vshotr,par%sgx(:,jsources),F)
    if(.not. allocated(bwd%Vshot))  allocate(bwd%Vshot, mold=fwd%Vshot)
    bwd%Vshot=fwd%Vshot
    if(.not. allocated(bwd%Vshotr)) allocate(bwd%Vshotr,mold=fwd%Vshotr)
    bwd%Vshotr=fwd%Vshotr

    !! Compute shot position and report progress
    fwd%shot_pos = 1+floor((sx(par%trace_shot)-gx(par%trace_shot))/par%dx)
    bwd%shot_pos = 1+floor((sx(par%trace_shot)-gx(par%trace_shot))/par%dx)
    write(0,'(a,i7,i7,i7,i7,i7,i7,i5)')     &
    'No. sources,curr source,shot pos.,source_index,first_geof,last_geof,ngeof',  &
    par%nsources,jsources,fwd%shot_pos,source_index(jsources,2),&
    par%first_geof,par%last_geof,par%ngeof
  
    !! Initialize updated velocity and wavefield classes

    select case(op) 
      case('W') 
        !! Create reflectivity frames for scattered wavefields
        call ref%rshot(par,wemva%pertshot,par%sgx(:,jsources))
  
        !! Reading data (traces from one shot gather)
        call par%read_shot()
  
        !! Perform WEMVA modeling for each shot gather
        call wemva%wemva(par,fwd,bwd,vel,ref,jsources)
  
        !! Build the wemva perturbation
        associate(x1=>1-par%napL,x2=>par%ngx+par%napR)
          call wemva%build_perturb(par,par%sgx(x1:x2,jsources))
        end associate
        dapprox=wemva%IPert
      case('B')
        !! Perform Born modeling
        call born%born(par,fwd,vel,ref,jsources)
      case default
        call skull()
        error stop "Linear_test(3): Unknown operator" 
    end select

    !! Perform forward operator on unperturbed model 
    select case(op) 
      case('W') 
        !! Perform Reverse-Time Migration in each shot gather
        call image%rtm(par,fwd,bwd,jsources)
  
        !! Build the migrated image
        associate(x1=>1-par%napL,x2=>par%ngx+par%napR)
          call image%build_images(par,par%sgx(x1:x2,jsources),F)
        end associate
  
        !! Temporally stores rtm image
        if(par%eic) then
          allocate(temp1,source=image%IRefl)
        else
          allocate(temp1,source=image%ICorr)
        end if
      case('B')
        par%src=src_tmp
        !! Perform non-linear modeling
        call nonlin%nonlin(par,fwd,vel,jsources)

        !! Temporally stores modeled data
        allocate(temp1,source=nonlin%dsynt)
      case default
        call skull()
        error stop "Linear_test(3): Unknown operator" 
    end select

    par%V_int=par%V_int+dV
    !! Create velocity frames for source wavefield propagation
    call vel%vshot(par,fwd%Vshot, par%sgx(:,jsources),T)
    fwd%Vshotr=fwd%Vshot
!    call vel%vshot(par,fwd%Vshotr,par%sgx(:,jsources),F)
    if(.not. allocated(bwd%Vshot))  allocate(bwd%Vshot, mold=fwd%Vshot)
    bwd%Vshot=fwd%Vshot
    if(.not. allocated(bwd%Vshotr)) allocate(bwd%Vshotr,mold=fwd%Vshotr)
    bwd%Vshotr=fwd%Vshotr

    !! Perform forward operator on perturbed model
    select case(op) 
      case('W') 
        !! Perform Reverse-Time Migration in each shot gather
        image%ICorr=0.
        image%IRefl=0.
        call image%rtm(par,fwd,bwd,jsources)
  
        !! Build the migrated image
        associate(x1=>1-par%napL,x2=>par%ngx+par%napR)
          call image%build_images(par,par%sgx(x1:x2,jsources),F)
        end associate
  
        if(par%eic) then
          image%IRefl=image%IRefl-temp1
          dexact =image%IRefl
        else
          image%ICorr=image%ICorr-temp1
          dexact =image%ICorr
        end if
      case('B')
        !! Perform non-linear modeling
        call nonlin%nonlin(par,fwd,vel,jsources)
  
        !! Subtract direct waves from non-linear modeled data
        nonlin%dsynt=nonlin%dsynt-temp1
  
        dapprox=born%dsynt
        dexact =nonlin%dsynt
  
        !! Write the images history files and binaries
        allocate(born%resd,source=dexact)
        call outp%write_history(par,born,jsources)
        call outp%write_history(par,born,jsources,.true.)
      case default
        call skull()
        error stop "Linear_test(3): Unknown operator" 
    end select

    !! De-allocate temporary array 
    deallocate(temp1)

write(0,*) 'norm2(dexact): ',norm2(dexact)
write(0,*) 'norm2(dapprox): ',norm2(dapprox)
    !! Advance to the next receiver and the corresponding traces
    par%trace_shot = par%trace_shot+par%ngeof
    par%first_geof = par%last_geof+1
  end do loop_over_sources
  write(0,*) ''
  write(0,*) '********************** MAIN LOOP FINISHES! ***********************'
  write(0,*) ''

  !! Write the images history files and binaries
  if(op=='W') then
    call outp%write_history(par,image,.false.)
    call outp%write_history(par,image,wemva)
  end if  

  !! Perform the linearity test
  write(0,*) 'Perform the linear test:'
  allocate(d1(size(dexact, 1)*size(dexact, 2)),&
           d2(size(dapprox,1)*size(dapprox,2)),&
           diff(size(dapprox,1)*size(dapprox,2)))
  d1=reshape(dexact, (/size(dexact, 1)*size(dexact, 2)/))
  d2=reshape(dapprox,(/size(dapprox,1)*size(dapprox,2)/))
  diff=d1-d2
  err_abs=dot_product(diff,diff)
  err_rel=err_abs/dot_product(d1,d1)
  write(0,*) 'err_abs: ',err_abs
  write(0,*) 'err_rel: ',err_rel

  
  !! Get the timing
  call date_and_time(values=timer1)
  timer0 = timer1-timer0
  timer_sum = timer_sum + (((timer0(3)*24+timer0(5))*60+timer0(6))*60+timer0(7))*1000+timer0(8)
  call to_history("timer sum (sec)", timer_sum/1000)
  call to_history("timer sum (min)", timer_sum/1000/60)
  write(0,*) "timer sum (sec)", timer_sum/1000
  write(0,*) "timer sum (min)", timer_sum/1000/60
  
  !! Finale!
  write(0,*) ' '
  write(0,*) 'Finishing the Linear_test program normally!'

end program
