program born2d
  !!**************************************************************
  !!*** Main program for 2D Born Modeling.                     ***
  !!*** Using finite differences with Lax-Wendroff method      ***
  !!*** Based upon rtm2d code rewritten in Modern Fortran      ***
  !!*** Author: Alejandro Cabrales                             ***
  !!*** Stanford Exploration Project/Petroleos Mexicanos, 2017 ***
  !!**************************************************************
  use sep
  use declare_variables_mod
  use read_mod, only : read_
  use write_mod, only : write_
  use src_rcv_interp_mod
  use vshot_mod, only : vshot_foo
  use rshot_mod, only : rshot_foo
  use fwd_propagation_mod, only : src_prop_fwd
  use born_mod, only : born_foo
  use taper_mod
  use skull_mod
  use assertions
  implicit none
  !! The following are type bound variables
  type(src_prop_fwd) :: fwd ! Arrays and procedures for source forward propagation
  type(born_foo)     :: born ! Arrays and procedures for imaging conditions
  type(read_)        :: par  ! Contains the input parameters and reading procedures
  type(write_)       :: outp ! Contains the ouput parameters and reading procedures
  type(vshot_foo)    :: vel
  type(rshot_foo)    :: ref
  
  write(0,'(a)') ' '
  write(0,'(a)') '************************************************************'
  write(0,'(a)') '*** 2D Born Modeling using FD with Lax-Wendroff method     *'
  write(0,'(a)') '*** Author: Alejandro Cabrales                             *'
  write(0,'(a)') '*** Stanford Exploration Project/Petroleos Mexicanos, 2017 *'
  write(0,'(a)') '************************************************************'
  write(0,'(a)') ' '
  
  !! Initialize sep libraries
  call sep_init()
  call init_3d()
  
  !! Initialize timing
  call DATE_and_TIME(values=timer0)
  
  !! Read the input history file
  call par%read_parameters()
  
  !! Read data history
  call par%read_history()
  
  !! Read data header
  call par%read_headers()
  
  !! Read velocity
  call par%read_velocity()
  
  !! Read reflectivity
  call par%read_reflectivity()
  
  !! Determine time sampling according to Courant-Fredrich-Lewis condition
  call par%courant_stuff()
  
  !! Obtain data geometry and apertures from input parameters
  call par%data_geometry()
  
  !! Read and verbose source
  call par%read_src()
  
  !! Verbose some useful parameters
  call par%verbose_par()
  
  !! Initialize parameters for forward propagations
  call fwd%fwd_prop_init(par)
  
  !! Initialize tapering for receiver propagation
  call taper_init(par,use_taper)
  
  !! Initialize derivatives coeffiecients
  call fwd%lap_init()
  
  !! Initialize variables for shot velocities
  call vel%vshot_init(par) 
  
  !! Initialize variables for reflectivity
  call ref%rshot_init(par) 
  
  !! Initialize parameters for born modeling
  call born%born_init(par,fwd,sx,gx)

  !! Initialize some variables for the main loop
  par%first_geof = 1
  par%ntSnap=0
  
  !! Initialize writing module array
  call outp%write_init(par) 
  
  !! This is the main loop to migrate the shots
  write(0,*) ''
  write(0,*) '********************** MAIN LOOP BEGINS! ***********************'
  write(0,*) ''
  loop_over_sources: do jsources = par%first_source,par%last_source
    !! Reading data (traces from one shot gather)
    call par%read_shot()
  
    !! Get number of geophone traces in shot gather (in some datasets is variable)
    par%ngeof = source_index(jsources,1)
    par%last_geof = par%first_geof+par%ngeof-1
  
    !! Create velocity frames for source wavefield propagation
    call vel%vshot(par,fwd%Vshotr,par%sgx(:,jsources),.true.)
    fwd%Vshot=fwd%Vshotr
    call sep_write(fwd%Vshot,'vrandom')
  
    !! Compute shot position and report progress
    fwd%shot_pos = 1+floor((sx(par%trace_shot)-gx(par%trace_shot))/par%dx)
    write(0,'(a,i7,i7,i7,i7,i7,i7,i5)')     &
    'No. sources,curr source,shot pos.,source_index,first_geof,last_geof,ngeof',  &
    par%nsources,jsources,fwd%shot_pos,source_index(jsources,2),&
    par%first_geof,par%last_geof,par%ngeof
  
    !! Perform Born Modeling for each shot gather
    call born%born(par,fwd,vel,ref,jsources)
  
    !! Write the images history files and binaries
    call outp%write_history(par,born,jsources)
  
    !! Advance to the next receiver and the corresponding traces
    par%trace_shot = par%trace_shot+par%ngeof
    par%first_geof = par%last_geof+1
  end do loop_over_sources
  write(0,*) ''
  write(0,*) '********************** MAIN LOOP FINISHES! ***********************'
  write(0,*) ''
  
  !! Get the timing
  call DATE_AND_TIME(values=timer1)
  timer0 = timer1-timer0
  timer_sum = timer_sum + (((timer0(3)*24+timer0(5))*60+timer0(6))*60+timer0(7))*1000+timer0(8)
  call to_history("timer sum (sec)", timer_sum/1000)
  call to_history("timer sum (min)", timer_sum/1000/60)
  write(0,*) "timer sum (sec)", timer_sum/1000
  write(0,*) "timer sum (min)", timer_sum/1000/60
  
  !! Finale!
  write(0,*) ' '
  write(0,*) 'Finishing the born2d program normally!'

end program
