module fwd_propagation_mod
  !! This module contains all procedures for computing 
  !! forward propagation of the source wavefield
  use sep
  use declare_variables_mod, only : timer0, timer1, timer_sum
  use interp1d_mod
  use skull_mod
  use read_mod, only : read_
  implicit none
  private
  public :: src_prop_fwd

  integer :: allocation_status
  integer :: i, jgx, jx, jz, jpol, jj, jd, jm, jaux
  integer, parameter :: buff=5
  real :: a0x, a1x, a2x, a3x, a4x, a5x, a6x, a0z, a1z, a2z, a3z, a4z, a5z, a6z
  real :: b1x, b2x, b3x, b4x, b5x, b6x, b1z, b2z, b3z, b4z, b5z, b6z
  real :: a0, idx, idz, dp_xx, dr_zz
  real :: fx, dt1, dt2, dt3, dt4
  real :: time0, time1
  real, dimension(:),     allocatable :: pauxgrad
  real, dimension(:,:),   allocatable :: psnap, Vel2
  real, dimension(:,:),   allocatable :: dpxz_act, dpx_cur, dpz_cur, dpxz_cur
  real, dimension(:,:,:), allocatable :: dpxz, drxz

  !! Pointers are needed to enter values to the C procedures for ISPC
  real, target,  dimension(:), allocatable:: pcur, pxcur, pzcur, rzcur
  real, target,  dimension(:), allocatable:: pnew, pxnew, pznew
  real, pointer, dimension(:) :: ppcur, ppxcur, ppzcur
  real, pointer, dimension(:) :: ppnew, ppxnew, ppznew

  type(read_), allocatable :: par
 
  type src_prop_fwd
    !! Type src_prop_fwd for source wavefield forward propagation
    integer :: shot_pos
    real, dimension(:),   allocatable :: src_scl
    real, dimension(:,:), allocatable :: rcv_scl, Vshot, Vshotr
    real, dimension(:,:), allocatable :: pftm1, pftm2, pft, pftX, pftZ
    real, dimension(:,:), allocatable :: dpftm1, dpftm2, dpft
    real, dimension(:,:), allocatable :: d2pftm1, d2pftm2, d2pft
    logical :: use_grad = .false.
  contains
    procedure fwd_prop_init
    procedure grad_init
    procedure grad
    procedure lap_init
    procedure lap
    procedure init_src_wavefields
!    procedure scale_src
!    procedure scale_rcv
    procedure fwd_prop
    procedure get_snaps_src
    procedure get_snaps_scat
  end type

  interface
    pure module subroutine fwd_prop_init(this,par_in)
      !! This procedure initializes all the scalar variables
      implicit none
      class(src_prop_fwd), intent(inout) :: this
      class(read_), intent(in) :: par_in
    end subroutine

    pure module subroutine grad_init(this)
      !! This procedure initializes first order derivative arrays
      implicit none 
      class(src_prop_fwd), intent(inout) :: this
    end subroutine
  
    pure module subroutine lap_init(this)
      !! This procedure initializes laplacian-related arrays
      implicit none 
      class(src_prop_fwd), intent(inout) :: this
    end subroutine
  
    pure module subroutine grad(this)
      !! This procedure calculates spatial first derivatives of wavefields
      implicit none
      class(src_prop_fwd), intent(inout) :: this
    end subroutine
  
    pure module subroutine lap(this,dpxz,isSrc,adj)
      !! This procedure obtains the laplacian of the wavefield
      implicit none
      class(src_prop_fwd),intent(in) :: this
      real, intent(inout) :: dpxz(:,:,:)
      logical, intent(in) :: isSrc, adj
    end subroutine

    pure module subroutine init_src_wavefields(this)
      !! This procedure sets to zero soource-wavefield-related arrays
      implicit none
      class(src_prop_fwd), intent(inout)  :: this
    end subroutine

    pure module subroutine get_snaps_src(this,sgx)
      !! This procedure obtains snapshots of the source wavefield from the chosen shot gather
      implicit none
      class(src_prop_fwd), intent(in) :: this
      real, dimension(:), intent(in) :: sgx
    end subroutine
 
    pure module subroutine get_snaps_scat(this,sgx)
      !! This procedure obtains snapshots of the source wavefield from the chosen shot gather
      implicit none
      class(src_prop_fwd), intent(in) :: this
      real, dimension(:), intent(in) :: sgx
    end subroutine
 
    pure module subroutine fwd_prop(this,adj)
      !! This procedure performs the forward propagation using the Lax-Wendroff method
      !! Here I'm testing the Object Pattern
      implicit none
      class(src_prop_fwd), intent(inout) :: this
      logical, intent(in) :: adj
    end subroutine
  end interface
end module
