submodule(grad_mod) grad_mod_implementation
  implicit none

contains
  module procedure grad_init
    associate(nkz=>par%nkz,nkgx=>par%nkgx,nz=>par%nz,nx=>par%nx)
    associate(dz=>par%dz,dx=>par%dx)
      idx = 1.0/dx
      idz = 1.0/dz
  
      allocate(dpx_cur(nkz,nkgx),&
               dpz_cur(nkz,nkgx),&
               pauxgrad(max(nkz,nkgx)),&
               pxcur(nkz*nkgx),&
               rzcur(nkz*nkgx),&
               stat=allocation_status)
      if(allocation_status /= 0) then
        call skull()
        error stop 'grad_init: could not allocate arrays!'
      end if
  
      if(par%ispc) then
        allocate(pxnew(nkz*nkgx),stat=allocation_status)
        if(allocation_status /= 0) then
          call skull()
          error stop 'grad_init: could not allocate array for ispc!'
        end if
      end if
    end associate
    end associate

    !! 1st deriv, 12th order approx. table by Fornberg and Ghrist (1999)
    b1x = 160083.0/131072.0 
    b2x = -12705.0/131072.0
    b3x = 22869.0/1310720.0
    b4x = -5445.0/1835008.0
    b5x = 847.0/2359296.0
    b6x = -63.0/2883584.0
    b1z = b1x
    b2z = b2x
    b3z = b3x
    b4z = b4x
    b5z = b5x
    b6z = b6x
    b1x = b1x*idx
    b2x = b2x*idx
    b3x = b3x*idx
    b4x = b4x*idx
    b5x = b5x*idx
    b6x = b6x*idx
    b1z = b1z*idz
    b2z = b2z*idz
    b3z = b3z*idz
    b4z = b4z*idz
    b5z = b5z*idz
    b6z = b6z*idz
  end procedure

  module procedure grad
    dpx_cur = bwd%pftX
    dpz_cur = bwd%pftZ
    bwd%pftX = 0
    bwd%pftZ = 0
    ppxcur => pxcur
    ppxnew => pxnew
    ppzcur => rzcur
    associate(nkz=>par%nkz,nkgx=>par%nkgx,ispc=>par%ispc)
    associate(dz=>par%dz,dx=>par%dx)
    associate(n1=>nkz,n2=>nkgx,px=>bwd%pftX,rz=>bwd%pftZ)
      if(ispc) then ! Use ISPC
        pxcur = reshape(dpx_cur(1:n1,1:n2),(/n1*n2/))
        rzcur = reshape(dpz_cur(1:n1,1:n2),(/n1*n2/))
        !$omp parallel do default(shared) private(jgx) 
        do jgx = 6,nkgx-6
          call gradXceic(0,0,ppxcur,ppxnew,nkz,jgx-1,dx)
        end do
        !$omp end parallel do

        !$omp parallel do default(shared) private(jgx) 
        do jgx = 1,nkgx
          call gradZceic(0,0,ppzcur,ppznew,nkz,jgx-1,dz)
        end do
        !$omp end parallel do
        px = reshape(pxnew,(/n1,n2/)) 
      else ! Don't use ISPC
        !$omp parallel do default(shared) private(jgx,jz,jaux) 
        do jgx = 6,n2-6
!          do concurrent(jz = 1:n1)
          do jz = 1,n1
            jaux=(jgx-1)*n1+jz
            pxcur(jaux) = b1x*(dpx_cur(jz,jgx+1)-dpx_cur(jz,jgx  ))+&
                          b2x*(dpx_cur(jz,jgx+2)-dpx_cur(jz,jgx-1))+&
                          b3x*(dpx_cur(jz,jgx+3)-dpx_cur(jz,jgx-2))+&
                          b4x*(dpx_cur(jz,jgx+4)-dpx_cur(jz,jgx-3))+&
                          b5x*(dpx_cur(jz,jgx+5)-dpx_cur(jz,jgx-4))+&
                          b6x*(dpx_cur(jz,jgx+6)-dpx_cur(jz,jgx-5))
          end do
        end do
        !$omp end parallel do
      
        !$omp parallel do default(shared) private(jgx,jz,jaux) 
        do jgx = 1,n2
!          do concurrent(jz = 6:n1-6)
          do jz = 6,n1-6
            jaux=(jgx-1)*n1+jz
            rzcur(jaux) = b1z*(dpz_cur(jz+1,jgx)-dpz_cur(jz,jgx  ))+&
                          b2z*(dpz_cur(jz+2,jgx)-dpz_cur(jz-1,jgx))+&
                          b3z*(dpz_cur(jz+3,jgx)-dpz_cur(jz-2,jgx))+&
                          b4z*(dpz_cur(jz+4,jgx)-dpz_cur(jz-3,jgx))+&
                          b5z*(dpz_cur(jz+5,jgx)-dpz_cur(jz-4,jgx))+&
                          b6z*(dpz_cur(jz+6,jgx)-dpz_cur(jz-5,jgx))
          end do
        end do
        !$omp end parallel do
        px = reshape(pxcur,(/n1,n2/)) 
        rz = reshape(rzcur,(/n1,n2/)) 
      end if
      pauxgrad=0
      !$omp parallel do default(shared) private(pauxgrad,jgx) 
      do jgx = 1,n2
        call interp1d_lop2(.false., .false., rz(:,jgx),pauxgrad(1:n1))
        rz(:,jgx)=pauxgrad(1:n1)
      end do
      !$omp end parallel do

      pauxgrad=0
      !$omp parallel do default(shared) private(pauxgrad,jz) 
      do jz = 1,n1
        call interp1d_lop2(.false., .false., px(jz,:),pauxgrad(1:n2))
        px(jz,:)=pauxgrad(1:n2)
      end do
      !$omp end parallel do
    end associate
    end associate
    end associate
  end procedure
end submodule
