program lsqrtm2d
 !!*** Perform Linearized Waveform Inversion (least-squares RTM) in 2D
  use sep
  use declare_variables_mod
  use read_mod, only : read_
  use write_mod, only : write_
  use src_rcv_interp_mod
  use vshot_mod, only : vshot_foo
  use rshot_mod, only : rshot_foo
  use fwd_propagation_mod, only : src_prop_fwd
  use bwd_propagation_mod, only : src_rcv_prop_bwd
  use rtm_mod, only : rtm_image
  use born_mod, only : born_foo
  use born_op_mod
  use cgstep_mod
  use taper_mod
  use skull_mod
  use assertions
  use solver_mod
  use solver_smp1_mod

  implicit none
  integer                :: time(8), seed(2), allocation_status
  integer                :: nfilt, nreg
  real                   :: eps
  real, dimension(:), allocatable :: model, data_, resd, bc
  type(src_prop_fwd)     :: fwd ! Arrays and procedures for source forward propagation
  type(src_rcv_prop_bwd) :: bwd ! Arrays and procedures for source/receiver backward propagation
  type(born_foo)         :: born ! Arrays and procedures for imaging conditions
  type(rtm_image)        :: image ! Arrays and procedures for imaging conditions
  type(read_)            :: par  ! Contains the input parameters and reading procedures
  type(write_)           :: outp ! Contains the ouput parameters and reading procedures
  type(vshot_foo)        :: vel
  type(rshot_foo)        :: ref

  !! Initialize sep libraries
  call sep_init()
  call init_3d()
  
  !! Initialize timing
  call date_and_time(values=timer0)
  
  !! Read the input history file
  call par%read_parameters()

  !! De-activate energy imaging conditions for lsqrtm
  par%eic=F
  
  !! Read data history
  call par%read_history()
  
  !! Read data header
  call par%read_headers()
  
  !! Read velocity
  call par%read_velocity()

  !! Determine time sampling according to Courant-Fredrich-Lewis condition
  call par%courant_stuff()
  
  !! Obtain data geometry and apertures from input parameters
  call par%data_geometry()
  
  !! Read and verbose source
  call par%read_src()
  
  !! Verbose some useful parameters
  call par%verbose_par()
  
  !! Initialize parameters for forward and backward propagations
  call fwd%fwd_prop_init(par)
  call bwd%bwd_prop_init(par)
  
  !! Initialize first derivatives parameters
  !call fwd%grad_init()
  
  !! Initialize variables for shot velocities
  call vel%vshot_init(par) 
  
  !! Initialize variables for reflectivity
  call ref%rshot_init(par) 
  
  !! Initialize tapering for receiver propagation
  call taper_init(par,use_taper)
  
  !! Initialize derivatives coeffiecients
  call fwd%lap_init()
  
  !! Initialize parameters for born modeling
  call born%born_init(par,fwd,sx,gx)

  !! Initialize parameters for imaging conditions
  call image%rtm_init(par,fwd,bwd,sx,gx)
 
  !! Initialize parameters for born inversion
  call born_op_init1(par,image,born)
 
  !! Initialize some variables for the main loop
  par%first_geof = 1
  par%ntSnap=0
  
  !! Initialize writing module array
  call outp%write_init(par) 

  !! Initialize convolution filter 
  !call from_aux("filter", "n1", nfilt)
  !allocate(bc(nfilt))
  !call sep_read(bc, "filter")
  !call tconv1_init1(par%nt*par%nkgx,nfilt)
  !call tconv1_init2(bc)
  !nreg=nfilt+par%nt*par%nkgx

  allocate(model(par%nkgx*par%nkz),&
           data_(par%nkgx*par%nt), &
            resd(par%nkgx*par%nt), &
           stat=allocation_status)
  if(allocation_status /= 0) then
    call skull()
    error stop 'lsqrtm2d: could not allocate model and data!'
  end if

  allocate(born%resd,mold=born%dsynt,stat=allocation_status)
  if(allocation_status /= 0) then
    call skull()
    error stop 'lsqrtm2d: could not allocate resd!'
  end if

!  allocate(err(par%niter),&
!           mmov(par%nkz*par%nkgx,par%niter),&
!           rmov(par%nt*par%nkgx,par%niter), &
!           stat=allocation_status)
!  if(allocation_status /= 0) then
!    call skull()
!    error stop 'lsqrtm2d: could not allocate err, mmov, rmov'
!  end if

  !! This is the main loop to migrate the shots
  write(0,*) ''
  write(0,*) '********************** MAIN LOOP BEGINS! ***********************'
  write(0,*) ''
  loop_over_sources: do jsources = par%first_source,par%last_source
    !! Get number of geophone traces in shot gather (in some datasets is variable)
    par%ngeof = source_index(jsources,1)
    par%last_geof = par%first_geof+par%ngeof-1
  
    !! Create velocity frames for source wavefield propagation
    call vel%vshot(par,fwd%Vshot,par%sgx(:,jsources), .true.)
    call vel%vshot(par,fwd%Vshotr,par%sgx(:,jsources),.false.)
    call sep_write(fwd%Vshot,'vrandom')
!    if(.not. allocated(bwd%Vshot))  allocate(bwd%Vshot, source=fwd%Vshot)
!    if(.not. allocated(bwd%Vshotr)) allocate(bwd%Vshotr,source=fwd%Vshotr)
    if(.not. allocated(bwd%Vshot))  allocate(bwd%Vshot, mold=fwd%Vshot)
    bwd%Vshot=fwd%Vshot
    if(.not. allocated(bwd%Vshotr)) allocate(bwd%Vshotr,mold=fwd%Vshotr)
    bwd%Vshotr=fwd%Vshotr

    !! Compute shot position and report progress
    fwd%shot_pos = 1+floor((sx(par%trace_shot)-gx(par%trace_shot))/par%dx)
    bwd%shot_pos = 1+floor((sx(par%trace_shot)-gx(par%trace_shot))/par%dx)
    write(0,'(a,i7,i7,i7,i7,i7,i7,i5)')     &
    'No. sources,curr source,shot pos.,source_index,first_geof,last_geof,ngeof', &
    par%nsources,jsources,fwd%shot_pos,source_index(jsources,2),&
    par%first_geof,par%last_geof,par%ngeof
  
    !! Reading data (traces from one shot gather)
    call par%read_shot()
    data_=reshape(par%rcv,(/par%nkgx*par%nt/))
    model=0.

    !! Initialize updated velocity and wavefield classes
    call born_op_init2(vel,ref,fwd,bwd,jsources)

    write(0,*) 'Begin linear solver'
    call solver_smp(model,data_,born_op,cgstep,par%niter,resd=resd,verb=.true.)
!                    resd=resd,mmov=mmov,rmov=rmov,err=err,verb=.true.)
!    call solver_reg1(model,data_,born_op,tconv1_lop,cgstep,nreg,par%niter,&
!                     eps=1e2,resd=resd,verb=.true.)
    write(0,*) 'Finish linear solver'

    image%num=reshape(model,(/par%nkz,par%nkgx/))

    !! Construct the images
    associate(x1=>1-par%napL,x2=>par%ngx+par%napR)
      call image%build_images(par,par%sgx(x1:x2,jsources),F)
    end associate

    !! Write the images history files and binaries
    born%resd=reshape(resd,(/par%nt,par%nkgx/))
    call outp%write_history(par,born,jsources,.true.)

    !! De-allocate objects in born_op module
    call born_op_deallocate()

    !! Advance to the next receiver and the corresponding traces
    par%trace_shot = par%trace_shot+par%ngeof
    par%first_geof = par%last_geof+1
  end do loop_over_sources
  write(0,*) ''
  write(0,*) '********************** MAIN LOOP FINISHES! ***********************'
  write(0,*) ''

  !! Write the images history files and binaries
  call outp%write_history(par,image,.false.)
  
  !! Get the timing
  call DATE_AND_TIME(values=timer1)
  timer0 = timer1-timer0
  timer_sum = timer_sum + (((timer0(3)*24+timer0(5))*60+timer0(6))*60+timer0(7))*1000+timer0(8)
  call to_history("timer sum (sec)", timer_sum/1000)
  call to_history("timer sum (min)", timer_sum/1000/60)
  write(0,*) "timer sum (sec)", timer_sum/1000
  write(0,*) "timer sum (min)", timer_sum/1000/60
  
  !! Finale!
  write(0,*) ' '
  write(0,*) 'Finishing the lsqrtm2d program normally!'

end program
