program lsqrtmHess2d
 !!*** Perform Linearized Waveform Inversion (least-squares RTM) in 2D
  use sep
  use declare_variables_mod
  use read_mod, only : read_
  use write_mod, only : write_
  use rtm_mod, only : rtm_image
  use H_mod
  use cgstep_mod
  use taper_mod
  use skull_mod
  use assertions
  use solver_mod
  use solver_smp1_mod
  implicit none
  integer                :: time(8), seed(2), allocation_status, stat
  integer                :: nfilt, nreg
  real                   :: eps
  real, dimension(:), allocatable :: model, data_, resd, bc
  real, dimension(:,:), allocatable :: diagH
  type(rtm_image)        :: image ! Arrays and procedures for imaging conditions
  type(read_)            :: par  ! Contains the input parameters and reading procedures
  type(write_)           :: outp ! Contains the ouput parameters and reading procedures

  !! Initialize sep libraries
  call sep_init()
  call init_3d()
  
  !! Initialize timing
  call date_and_time(values=timer0)
  
  !! Read the input history file
  call par%read_parameters()

  !! De-activate energy imaging conditions for lsqrtmHess
  par%eic=F
  
  !! Read data history
  call par%read_history()
  
  !! Read data header
  call par%read_headers()
  
  !! Read velocity
  call par%read_velocity()

  !! Read "blurred" reflectivity (RTM image)
  call par%read_reflectivity()
  
  !! Determine time sampling according to Courant-Fredrich-Lewis condition
  call par%courant_stuff()
  
  !! Obtain data geometry and apertures from input parameters
  call par%data_geometry()
  
  !! Read and verbose source
  call par%read_src()
  
  !! Verbose some useful parameters
  call par%verbose_par()
  

  !! Initialize parameters for imaging conditions
!  call image%rtm_init(par,fwd,bwd,sx,gx)
  allocate(image%ICorr(par%nz,par%first_cdp:par%last_cdp))
  allocate(diagH(par%nz,par%nx))
  !! Initialize parameters for GN Hessian inversion
  call H_init(par,T,diagH)

  !! Initialize writing module array
  call outp%write_init(par) 

  !! Initialize convolutional filter 
!  call from_aux("filter", "n1", nfilt)
!  allocate(bc(nfilt))
!  call sep_read(bc, "filter")
!  call tconv1_init1(par%nt*par%nkgx,nfilt)
!  call tconv1_init2(bc)
!  nreg=nfilt+par%nt*par%nkgx

  allocate(model(par%nx*par%nz),&
           data_(par%nx*par%nz), &
            resd(par%nx*par%nz), &
           stat=allocation_status)
  if(allocation_status /= 0) then
    call skull()
    error stop 'lsqrtm2d: could not allocate model and data!'
  end if

  !! This is the main loop to migrate the shots
  write(0,*) ''
  write(0,*) '********************** MAIN LOOP BEGINS! ***********************'
  write(0,*) ''
  loop_over_sources: do jsources = par%first_source,par%last_source
    !! Compute shot position and report progress
    write(0,'(a,i7,i7)') 'No. sources,curr source',par%nsources,jsources
  
    !! Reading data (traces from one shot gather)
    data_=reshape(par%refl,(/par%nx*par%nz/))
    model=0.

    if(par%nlagx==0 .and. par%nlagz==0) then
      write(0,*) 'No inversion; only divide by the diagonal of the Hessian'
      image%ICorr=par%refl/(diagH+0.001*sqrt(norm2(diagH)))
    else
      write(0,*) 'Begin linear solver'
      call solver_smp(model,data_,H_op,cgstep,par%niter,resd=resd,verb=.true.)
!                      mmov=mmov,rmov=rmov,err=err,verb=.true.)
!      call solver_reg1(model,data_,H_op,tconv1_lop,cgstep,nreg,par%niter,&
!                       eps=1e3,resd=resd,verb=.true.)

      write(0,*) 'Finish linear solver'
      image%ICorr=reshape(model,(/par%nz,par%nx/))
    end if
  end do loop_over_sources

  write(0,*) ''
  write(0,*) '********************** MAIN LOOP FINISHES! ***********************'
  write(0,*) ''

  !! Write the images history files and binaries
  call outp%write_history(par,image,F,"hess_inv_refl")
!  image%ICorr=reshape(resd,(/par%nz,par%nx/))
!  call outp%write_history(par,image,F,"resd")
  
  !! Get the timing
  call date_and_time(values=timer1)
  timer0 = timer1-timer0
  timer_sum = timer_sum + (((timer0(3)*24+timer0(5))*60+timer0(6))*60+timer0(7))*1000+timer0(8)
  call to_history("timer sum (sec)", timer_sum/1000)
  call to_history("timer sum (min)", timer_sum/1000/60)
  write(0,*) "timer sum (sec)", timer_sum/1000
  write(0,*) "timer sum (min)", timer_sum/1000/60
  
  !! Finale!
  write(0,*) ' '
  write(0,*) 'Finishing the lsqrtmHess2d program normally!'

end program
