program nonlin2d
  !!**************************************************************
  !!*** Main program for 2D Born Modeling.                     ***
  !!*** Using finite differences with Lax-Wendroff method      ***
  !!*** Based upon rtm2d code rewritten in Modern Fortran      ***
  !!*** Author: Alejandro Cabrales                             ***
  !!*** Stanford Exploration Project/Petroleos Mexicanos, 2017 ***
  !!**************************************************************
  use sep
  use declare_variables_mod
  use read_mod, only : read_
  use write_mod, only : write_
  use src_rcv_interp_mod
  use vshot_mod, only : vshot_foo
  use rshot_mod, only : rshot_foo
  use fwd_propagation_mod, only : src_prop_fwd
  use born_mod,  only : born_approx
  use nonlin_mod, only : nonlin_foo
  use taper_mod
  use skull_mod
  use assertions
  implicit none
  integer :: jgx, dummy
  integer, external :: sep_copy_hff
  real :: time0, time1
  real, allocatable :: Vtop(:)
  real, allocatable :: temp1(:,:)
  character(len=100) :: data_fn
  !! The following are type bound variables
  type(src_prop_fwd) :: fwd ! Arrays and procedures for source forward propagation
  type(nonlin_foo)   :: nonlin ! Arrays and procedures for non-linear modeling
  type(born_approx)  :: born ! Dummy object for writing purposes
  type(read_)        :: par  ! Contains the input parameters and reading procedures
  type(write_)       :: outp ! Contains the ouput parameters and reading procedures
  type(vshot_foo)    :: vel
  
  write(0,'(a)') ' '
  write(0,'(a)') '**************************************************************'
  write(0,'(a)') '*** 2D Non-linear Modeling using FD with Lax-Wendroff method *'
  write(0,'(a)') '*** Author: Alejandro Cabrales                               *'
  write(0,'(a)') '*** Stanford Exploration Project/Petroleos Mexicanos, 2017   *'
  write(0,'(a)') '**************************************************************'
  write(0,'(a)') ' '
  
  !! Initialize sep libraries
  call sep_init()
  call init_3d()
  
  !! Initialize timing
  call date_and_time(values=timer0)
  
  !! Read the input history file
  call par%read_parameters()

 ! call from_param('data',data_fn)

  !! Read data history
  call par%read_history()

  if(par%ntracesData==0) then
    !! Read from parameters file
     call par%par_geometry()
  else
    !! Read data header
    call par%read_headers()
  end if

  !! Read velocity
  call par%read_velocity()
  
  !! Determine time sampling according to Courant-Fredrich-Lewis condition
  call par%courant_stuff()

  !! Proceed when datafile is provided
  if(par%ntracesData/=0) then
    !! Obtain data geometry and apertures from input parameters
    call par%data_geometry()

!    if(par%dgxOdx==1) then
      !! Copy the headers from datafile to nldatafile
      dummy=sep_copy_hff("datafile","out")
      if(dummy < 0) & 
      call seperr('nonlin2d: Cannot copy headers from datafile to nldatafile')
!    else
      !! Copy the headers after interpolation
!      call outp%put_headers_data(par,data_fn)
!    end if
!  else
!    call par%put_headers_data()
  end if

  !! Read and verbose source
  call par%read_src(T)

  !! Verbose some useful parameters
  call par%verbose_par()
  
  !! Initialize parameters for forward propagations
  call fwd%fwd_prop_init(par)
  
  !! Initialize tapering for receiver propagation
  call taper_init(par,use_taper)
  
  !! Initialize derivatives coefficients
  call fwd%lap_init()
  
  !! Initialize variables for shot velocities
  call vel%vshot_init(par) 
  
  !! Initialize parameters for non-linear modeling
  call nonlin%nonlin_init(par,fwd,sx,gx)

  !! Initialize some variables for the main loop
  par%first_geof = 1
  par%ntSnap=0
 
  !! Initialize writing module array
  call outp%write_init(par)
  
  allocate(Vtop(par%nkgx))
  
  !! This is the main loop to model the shots
  write(0,*) ''
  write(0,*) '********************** MAIN LOOP BEGINS! ***********************'
  write(0,*) ''
  loop_over_sources: do jsources = par%first_source,par%last_source
  
    !! Get number of geophone traces in shot gather (in some datasets is variable)
    par%ngeof = source_index(jsources,1)
    par%last_geof = par%first_geof+par%ngeof-1
  
    !! Create velocity frames for source wavefield propagation
    call vel%vshot(par,fwd%Vshotr,par%sgx(:,jsources),F)
    call sep_write(fwd%Vshotr,'vrandom')
    fwd%Vshot=fwd%Vshotr
    Vtop=fwd%Vshot(1+par%zpad+par%pad_,:)

    !! Compute shot position and report progress
    fwd%shot_pos = 1+floor((sx(par%trace_shot)-gx(par%trace_shot))/par%dx)
    write(0,'(a,i7,i7,i7,i7,i7,i7,i5)')     &
    'No. sources,curr source,shot pos.,source_index,first_geof,last_geof,ngeof',  &
    par%nsources,jsources,fwd%shot_pos,source_index(jsources,2),&
    par%first_geof,par%last_geof,par%ngeof
  
    !! Perform Non-linear Modeling for each shot gather
    call nonlin%nonlin(par,fwd,vel,jsources)

    !! Turn snapshots off
    par%snapsht=.false.

    !! Temporally stores modeled data
    allocate(temp1,source=nonlin%dsynt)

    !! Prepare velocity at the surface for direct wave modeling
    do jgx=1,par%nkgx
      fwd%Vshot(:,jgx)=Vtop(jgx)
    end do

    !! Perform Non-linear Modeling for direct waves computation
    call nonlin%nonlin(par,fwd,vel,jsources)

    !! Subtract direct waves from non-linear modeled data
    nonlin%dsynt=temp1-nonlin%dsynt
  
    !! Turn snapshots on
    par%snapsht=.true.
    !! Write the images history files and binaries
    if(par%ntracesData==0) then
      call outp%write_history(par,nonlin,jsources)
    else
      call outp%write_history(par,born,nonlin,jsources)
    end if
 
    !! De-allocate temporary array 
    deallocate(temp1)

    !! Advance to the next receiver and the corresponding traces
    par%trace_shot = par%trace_shot+par%ngeof
    par%first_geof = par%last_geof+1
  write(0,*) ''
  write(0,*) '********************** MAIN LOOP FINISHES! ***********************'
  write(0,*) ''
  end do loop_over_sources


  !! Get the timing
  call date_and_time(values=timer1)
  timer0 = timer1-timer0
  timer_sum = timer_sum + (((timer0(3)*24+timer0(5))*60+timer0(6))*60+timer0(7))*1000+timer0(8)
  call to_history("timer sum (sec)", timer_sum/1000)
  call to_history("timer sum (min)", timer_sum/1000/60)
  write(0,*) "timer sum (sec)", timer_sum/1000
  write(0,*) "timer sum (min)", timer_sum/1000/60
  
  !! Finale!
  write(0,*) ' '
  write(0,*) 'Finishing the nonlin2d program normally!'

end program
