module read_mod
  !! This module performs reading/writing of command-line parameters, history files,
  !! header files, and binary files from segy-derived sep files
  use sep
  use declare_variables_mod, only : source_index, sx, gx
  use src_rcv_interp_mod
  use skull_mod
  use assertions, only : assert
  implicit none
  private
  public read_

  integer :: dummy, input_error, allocation_status
  integer :: k, l, jt, jx, jgx, jtraces, jsources
  integer :: ngx_p, nxpad, thetac
  integer :: scalco_key, sx_key, gx_key, sy_key, gy_key, n_keys_data, n_keys_vel
  integer :: next_header_out, previous_trace
  integer, external :: sseek_block, sreed
  integer, external :: sep_get_val_headers, sep_get_header_format_tag
  integer, external :: sep_get_number_keys, sep_get_key_index
  integer , dimension(:,:), allocatable :: headerI_in
  real, dimension(:,:), allocatable :: header_in
  real :: tIntFact, ox_shift, scalcof, frac_time
  real :: hsx_cur,hsx_next, hgx_cur, hgx_next, osx, ogx
  real :: z_start, z_end, hmin, fNyq, fc, jfx
  real, parameter :: UNDEFINED = -9999999.
  real, parameter :: pi = 4.*atan(1.) ! The Order of Chaos (Darren Aronofsky, 1998)
  real, parameter :: deg2rad = pi/180.
  real, allocatable, dimension(:)   :: hoff, srcData, rcvData, d2rcvData, sgx_max
  real, allocatable, dimension(:,:) :: vinput
  logical :: verb_src
  character :: char_aux
  character(len=100) :: wavelet_fn, tag_input, string_aux
  character(:), allocatable :: datafile, reflfile

  type read_
    !! This type bound contains input parameters and the corresponding procedures
    !! It is convenient to encapsulate them for easier handling
    integer :: ntData, nt, ntsData, nts, nz, nx, ngx, niter
    integer :: nbox, nkz, nkgx, ngeof, npts, nthreads
    integer :: first_cdp, last_cdp, first_source, last_source
    integer :: napL, napR, trace_shot, nsources, ntracesData, ntraces, jtStep
    integer :: first_geof, last_geof, jgeof, pad_, zpad, src_pad
    integer :: scalco, src_snap, ntSnap
    integer :: nlagz, nlagx, incPSFz, incPSFx
    real    :: dtData, dt, dz, dx, dsx, dgx, dgxOdx, o1, o2, d2, d3
    real    :: vel_fracc, Vmin, Vmax, Vcte, cfl, t_end
    real    :: apertureL, apertureR, cos2thc, alpha
    logical :: snapsht, one_src, ispc, eic, override_sx_header
    real, allocatable, dimension(:)     :: src
    real, allocatable, dimension(:,:)   :: rcv, V_int, refl, sgx
    real, allocatable, dimension(:,:,:) :: rcvs
    character(len=15)  :: unit1, unit2
!    character(len=:), allocatable :: projpref
    character(100) :: projpref
  contains
    procedure read_parameters
    procedure read_history
    procedure read_headers
    procedure read_velocity
    procedure read_reflectivity
    procedure courant_stuff
    procedure read_src
    procedure par_geometry
    procedure data_geometry
    procedure read_shot
    procedure verbose_par
  end type
  
  interface
    pure module subroutine read_parameters(this)
      !! This procedure reads parameters from the Makefile and the parameters
      !! file in folder "Par"
      implicit none
      class(read_), intent(inout) :: this
    end subroutine
  
    pure module subroutine read_history(this)
       ! This procedure reads parameters from history files, and calculates
       ! other parameters out of them
      implicit none
      class(read_), intent(inout) :: this
    end subroutine
  
    pure module subroutine read_headers(this)
      !! This procedure reads header data, and determines No. of traces, spacings, etc.
      implicit none
      class(read_), intent(inout) :: this
    end subroutine
  
    pure module subroutine read_velocity(this)
      !! This procedure reads velocity from the binary file (for Born Modeling)
      implicit none
      class(read_), intent(inout) :: this
    end subroutine
  
    pure module subroutine read_reflectivity(this)
      !! This procedure reads reflectivity from the binary file
      implicit none
      class(read_), intent(inout) :: this
    end subroutine
  
    pure module subroutine courant_stuff(this)
      !! This procedure determines the time step required to keep the
      !! computations stable, according to the Courant condition
      implicit none
      class(read_), intent(inout) :: this
    end subroutine
  
    pure module subroutine read_src(this,noderiv,wavelet_fn_add)
      !! This procedure reads the input wavelet and interpolated to the fine time step
      implicit none
      class(read_), intent(inout) :: this
      logical, optional, intent(in) :: noderiv
      character(100), optional, intent(in) :: wavelet_fn_add
    end subroutine
  
   pure  module subroutine par_geometry(this)
      !! This procedure creates the data geometry using information from the
      !! parameters file
      implicit none
      class(read_), intent(inout) :: this
    end subroutine
  
   pure  module subroutine data_geometry(this)
      !! This procedure creates the data geometry using information from the
      !! headers
      implicit none
      class(read_), intent(inout) :: this
    end subroutine
  
    pure module subroutine read_shot(this,dta2der)
      !! This procedure reads data traces and interpolate them to the fine time
      !! step and, if required, to a fine spatial grid
      implicit none
      class(read_), intent(inout) :: this
      logical, optional, intent(in) :: dta2der
    end subroutine

!    pure module subroutine read_shots(this)
!      !! This procedure reads all data traces of the shot at once
!      implicit none
!      class(read_), intent(inout) :: this
!    end subroutine

   pure module subroutine verbose_par(this)
     !! This procedure reports the parameters on the standard output
     implicit none
     class(read_), intent(in) :: this
   end subroutine
  end interface
end module
