submodule(rtm_mod) rtm_mod_implementation
  implicit none
contains
  module procedure rtm
    !! Initialize/update wavefield arrays
    call fwd%init_src_wavefields()
    call bwd%init_src_wavefields()
  
    !! Initialize numerators and denominators for imaging conditions
    call this%update_imag(par)

    !! Source scaling and receiver factors (using dipole sources)
    associate(z1=>par%zpad+par%pad_,x1=>par%pad_+par%napL+bwd%shot_pos)
      src_scl1=-(fwd%Vshot(1+z1,x1)*par%dt)**2
      src_scl2=-(fwd%Vshot(2+z1,x1)*par%dt)**2
    end associate
    if(.not. allocated(rcv_scl)) allocate(rcv_scl(par%nkz,par%nkgx))
    rcv_scl=-(fwd%Vshot(:,:)*par%dt)**2

    write(0,*) 'Perform reverse-time migration'

    do jt=1,nt+2
      !! Forward propagation of source wavefield
      if(jt>1) then
        call fwd%fwd_prop(F)
      end if

      !! Inject the source after scaling
      if(jt<=nt) then
        fwd%pft(1,bwd%shot_pos)=fwd%pft(1,bwd%shot_pos)+par%src(jt)*src_scl1
        fwd%pft(2,bwd%shot_pos)=fwd%pft(2,bwd%shot_pos)-par%src(jt)*src_scl2
      end if

      !! Obtain snapshots of the source wavefield for a single source
      if(par%snapsht .and. jsources == par%src_snap&
                     .and. mod(jt,jtsnap) == 0) then
        par%ntSnap=par%ntSnap+1
        call fwd%get_snaps_src(par%sgx(:,jsources))
      end if
    end do

    !! Copy the source wavefield frames to the reversible version
    call bwd%update_wavefields(fwd)

    do jt=nt,1,-1
      !! Inject the data
      bwd%prt(1,:)=bwd%prt(1,:)+par%rcv(jt,:)
      bwd%prt(2,:)=bwd%prt(2,:)-par%rcv(jt,:)

      !! Backward propagation of source wavefield
      call bwd%bwd_prop(fwd)   
      if(par%eic) call bwd%grad_rcv(fwd) 

      !! Re-inject the source after scaling
      if(jt<=nt-2) then
        bwd%pft(1,bwd%shot_pos)=bwd%pft(1,bwd%shot_pos)+par%src(jt+2)*src_scl1
        bwd%pft(2,bwd%shot_pos)=bwd%pft(2,bwd%shot_pos)-par%src(jt+2)*src_scl2
      end if

      !! Use damping function in receiver propagation wavefields
      if(use_taper) call taperings(bwd%prtp2,bwd%prtp1,bwd%prt)
      if(use_taper) call taperings(bwd%dprtp2,bwd%dprtp1,bwd%dprt)
      if(use_taper) call taperings(bwd%prtX,bwd%prtZ,bwd%d2prt)

      !! DELETE AFTER TEST Obtain snapshots of the source wavefield for a single source
      if(par%snapsht .and. jsources == par%src_snap&
                     .and. mod(jt,jtsnap) == 0) then
        call bwd%get_snaps_src(par%sgx(:,jsources))
      end if

      !! Obtain snapshots of the source wavefield for a single source
      if(par%snapsht .and. jsources==par%src_snap .and. mod(jt,jtsnap)==0) &
        call bwd%get_snaps_rcv(par%sgx(:,jsources))

      !! Apply imaging condition after scaling receiver wavefield
      this%num=this%num+bwd%pft*bwd%prt*rcv_scl 
      this%den=this%den+bwd%pft*bwd%pft*rcv_scl 

      !! Apply energy imaging conditions for imaging only
      if(par%eic) then
        this%num0=this%num0+bwd%pftX*bwd%prtX+bwd%pftZ*bwd%prtZ&
                           +bwd%dpft*bwd%dprt/bwd%Vshotr**2
        this%num90=this%num90+bwd%pftX*bwd%prtX+bwd%pftZ*bwd%prtZ&
               +par%cos2thc*bwd%dpft*bwd%dprt/bwd%Vshotr**2
        this%den0=this%den0+bwd%pftX*bwd%pftX+bwd%pftZ*bwd%pftZ&
                           +bwd%dpft*bwd%dpft/bwd%Vshotr**2
      end if
    end do
  end procedure

  module procedure rtm_init
    zshift = par%pad_+par%zpad
    x1 = 1-par%napL
    x2 = par%ngx+par%napR
    x3 = x1-par%pad_
    x4 = x2+par%pad_
    z1 = 1-par%zpad-par%pad_
    z2 = par%nz+par%pad_
    z3 = 1+zshift
    z4 = par%nz+zshift
    nt=par%nt
    nx=par%nx
    ngx=par%ngx
    nz=par%nz
    first_cdp=par%first_cdp
    last_cdp=par%last_cdp

    allocate(this%num  (z1:z2,x3:x4),&
             this%num0 (z1:z2,x3:x4),&
             this%num90(z1:z2,x3:x4),&
             this%den  (z1:z2,x3:x4),&
             this%den0 (z1:z2,x3:x4),&
             this%ICorr(nz,first_cdp:last_cdp),&
             this%ITomo(nz,first_cdp:last_cdp),&
             this%IRefl(nz,first_cdp:last_cdp),&
             stat=allocation_status)
    if(allocation_status /= 0) then
      call skull()
      error stop 'rtm_init: could not allocate arrays!'
    end if
    this%ICorr=0
    this%ITomo=0
    this%IRefl=0

    allocate(owt(par%ngx),stat=allocation_status)
    if(allocation_status /= 0) then
      call skull()
      error stop 'rtm_init: could not allocate owt array!'
    end if

    ! This is a tapering function with offset, employed by Ali (disabled now)
    do jgx=1,par%ngx
      owt(jgx) = -1.*abs(sx(par%trace_shot)-gx(jgx))
    end do
    owt = owt - minval(owt) + 1.
    owt = sin(pi/2.*owt/maxval(owt))*sin(pi/2.*owt/maxval(owt)) 
  end procedure

  module procedure update_imag
    this%num  =0.
    this%num0 =0.
    this%num90=0.
    this%den  =0.
    this%den0 =0.
  end procedure

  module procedure xcorrIC
    !$omp parallel do default(shared) private(jgx)
    loop_crosscorrelation:do jgx = x1,x2
      this%num(:,jgx)=this%num(:,jgx)+bwd%pft(1:nz,jgx)*bwd%prt(1:nz,jgx) 
!      this%num(:,jgx)=this%num(:,jgx)+prt_aux(1:nz,jgx)*bwd%prt(1:nz,jgx) 
      this%den(:,jgx)=1. !this%den(:,jgx)+bwd%pft(1:nz,jgx)*bwd%pft(1:nz,jgx)
    end do loop_crosscorrelation
    !$omp end parallel do

    !! Test for zero in den
    call assert(norm2(this%den)/=0,"xcorrIC: Zero denominators in den")
  end procedure

  module procedure energyIC
    !! Obtain the gradient components of the source and receiver wavefields
    bwd%pftX=bwd%prt
    bwd%pftZ=bwd%prt
    call bwd%grad() 
    bwd%prtX=bwd%pftX
    bwd%prtZ=bwd%pftZ
  
    bwd%pftX=bwd%pft
    bwd%pftZ=bwd%pft
    call bwd%grad() 
  
    k=0
    !$omp parallel do default(shared) private(jgx,k)
    loop_energy:do jgx = x1,x2
      k=jgx-x1+1+par%pad_
!      k=par%sgx(jgx,jsources)
      this%num0(:,jgx) =this%num0(:,jgx) +bwd%pftX(1:nz,jgx)*bwd%prtX(1:nz,jgx)&
                                         +bwd%pftZ(1:nz,jgx)*bwd%prtZ(1:nz,jgx)&
                                         +bwd%dpft(1:nz,jgx)*bwd%dprt(1:nz,jgx)&
                                         /bwd%Vshotr(z3:z4,k)**2

      this%num90(:,jgx)=this%num90(:,jgx)+bwd%pftX(1:nz,jgx)*bwd%prtX(1:nz,jgx)&
                                         +bwd%pftZ(1:nz,jgx)*bwd%prtZ(1:nz,jgx)&
                             +par%cos2thc*bwd%dpft(1:nz,jgx)*bwd%dprt(1:nz,jgx)&
                                         /bwd%Vshotr(z3:z4,k)**2

      this%den0(:,jgx) =this%den0(:,jgx) +bwd%pftX(1:nz,jgx)*bwd%pftX(1:nz,jgx)&
                                         +bwd%pftZ(1:nz,jgx)*bwd%pftZ(1:nz,jgx)&
                                         +bwd%dpft(1:nz,jgx)*bwd%dpft(1:nz,jgx)&
                                         /bwd%Vshotr(z3:z4,k)**2
    end do loop_energy
    !$omp end parallel do
    !! Test for zero in den0
    call assert(norm2(this%den0)/=0,"buid_images: Zero denominators in den0")
  end procedure

  module procedure build_images
    if(conv) then
      !! 2D Triangle filter according to Guitton et al. (2006)
      dummy=base_tris2d_oper(this%den,par%npts,par%npts,10) 
      tinyden=tiny_*sum(this%den)/size(this%den)
      dummy=base_tris2d_oper(this%den0,par%npts,par%npts,10) 
      tinyden0=tiny_*sum(this%den0)/size(this%den0)
    else
      tinyden =0.
      tinyden0=0.
      this%den =1.
      this%den0=1.
    end if

    k=0
    associate(IC=>this%ICorr,IT=>this%ITomo,IR=>this%IRefl)
    associate(n=>this%num,n0=>this%num0,n90=>this%num90,d=>this%den,d0=>this%den0)
      !$omp parallel do default(shared) private(jgx,k,jx,fx)
      do jgx = x1,x2
        k=jgx-x1+1
        fx = nint(1.5001+(sgx(k)-0.0)/par%dx)
        jx = fx
        fx = fx-jx 
        if(jx >= first_cdp .and. jx < last_cdp) then
          IC(:,jx)=IC(:,jx)+(1-fx)*n(1:nz,jgx)  /(d(1:nz,jgx)  +tinyden)&
                           +   fx *n(1:nz,jgx+1)/(d(1:nz,jgx+1)+tinyden)
          !! Test for NaNs
          call assert(.not. any(isnan(IC)),&
                      "build_images: NaNs detected in ICorr")
        end if
      end do
      !$omp end parallel do

      if(par%eic) then
        !$omp parallel do default(shared) private(jgx,k,jx,fx)
        do jgx = x1,x2
          k=jgx-x1+1
          fx = nint(1.5001+(sgx(k)-0.0)/par%dx)
          jx = fx
          fx = fx-jx 
          if(jx >= first_cdp .and. jx < last_cdp) then
            IT(:,jx)=IT(:,jx)+(1-fx)*n0(1:nz,jgx)  /(d0(1:nz,jgx)  +tinyden0)&
                             +   fx *n0(1:nz,jgx+1)/(d0(1:nz,jgx+1)+tinyden0)
  
            IR(:,jx)=IR(:,jx)+(1-fx)*n90(1:nz,jgx)  /(d0(1:nz,jgx)  +tinyden0)&
                             +   fx *n90(1:nz,jgx+1)/(d0(1:nz,jgx+1)+tinyden0)
  
            !! Test for NaNs
            call assert(.not. any(isnan(IT)),&
                        "build_images: NaNs detected in Itomo")
            call assert(.not. any(isnan(IR)),&
                        "build_images: NaNs detected in Irefl")
          end if
        end do
        !$omp end parallel do
      end if
    end associate
    end associate
  end procedure
end submodule
