program wemva2d
  !!*******************************************************************************
  !!*** Main program for 2D Wave-Equation Velocity Analysis (WEMVA) modeling.   ***
  !!*** Using finite differences with Lax-Wendroff method                       ***
  !!*** Author: Alejandro Cabrales                                              ***
  !!*** Stanford Exploration Project/Petroleos Mexicanos, 2017                  ***
  !!*******************************************************************************
  use sep
  use declare_variables_mod
  use read_mod, only : read_
  use write_mod, only : write_
  use src_rcv_interp_mod
  use vshot_mod, only : vshot_foo
  use rshot_mod, only : rshot_foo
  use fwd_propagation_mod, only : src_prop_fwd
  use bwd_propagation_mod, only : src_rcv_prop_bwd
  use rtm_mod, only : rtm_image
  use wemva_mod, only : wemva_foo
  use wemva_op_mod
  use taper_mod
  use skull_mod
  use assertions
  implicit none
  integer :: stat
  character :: char_aux
  logical :: adj
  !! The following are type bound variables
  type(src_prop_fwd) :: fwd     ! Arrays and procedures for source forward propagation
  type(src_rcv_prop_bwd) :: bwd ! Arrays and procedures for source/receiver backward propagation/
  type(rtm_image)        :: image ! Arrays and procedures for imaging conditions
  type(wemva_foo)    :: wemva   ! Arrays and procedures for imaging conditions
  type(read_)        :: par     ! Contains the input parameters and reading procedures
  type(write_)       :: outp    ! Contains the ouput parameters and reading procedures
  type(vshot_foo)    :: vel
  type(rshot_foo)    :: ref
  
  write(0,'(a)') ' '
  write(0,'(a)') '************************************************************'
  write(0,'(a)') '*** 2D Wave-Equation Velocity Analysis (WEMVA) Modeling    *'
  write(0,'(a)') '*** using FD with Lax-Wendroff method                      *'
  write(0,'(a)') '*** Author: Alejandro Cabrales                             *'
  write(0,'(a)') '*** Stanford Exploration Project/Petroleos Mexicanos, 2017 *'
  write(0,'(a)') '************************************************************'
  write(0,'(a)') ' '
  
  !! Initialize sep libraries
  call sep_init()
  call init_3d()
  
  !! Initialize timing
  call date_and_time(values=timer0)
  
  !! Read the input history file
  call par%read_parameters()
  par%eic=F

  !! Forward or adjoint?
  call from_param('adj',char_aux,'n')
    adj = (char_aux == 'y')

  !! Read data history
  call par%read_history()
  
  !! Read data header
  call par%read_headers()
  
  !! Read velocity
  call par%read_velocity()
  
  !! Read reflectivity
  call par%read_reflectivity()
  
  !! Determine time sampling according to Courant-Fredrich-Lewis condition
  call par%courant_stuff()
  
  !! Obtain data geometry and apertures from input parameters
  call par%data_geometry()
  
  !! Read and verbose source
  call par%read_src()
  
  !! Verbose some useful parameters
  call par%verbose_par()
  
  !! Initialize parameters for forward and backward propagations
  call fwd%fwd_prop_init(par)
  call bwd%bwd_prop_init(par)
  
  !! Initialize first derivatives parameters
  if(par%eic) call fwd%grad_init()
  if(par%eic) call bwd%grad_init()
  
  !! Initialize tapering for receiver propagation
  call taper_init(par,use_taper)
  
  !! Initialize derivatives coeffiecients
  call fwd%lap_init()
  
  !! Initialize variables for shot velocities
  call vel%vshot_init(par) 
  
  !! Initialize variables for reflectivity
  call ref%rshot_init(par) 
  
  !! Initialize parameters for wemva
  call wemva%wemva_init(par,fwd,sx,gx)

  !! Initialize wemva function (1)
  call wemva_op_init1(par,wemva)

  !! Initialize some variables for the main loop
  par%first_geof = 1
  par%ntSnap=0
  
  !! Initialize writing module array
  call outp%write_init(par) 
  
  !! This is the main loop to migrate the shots
  write(0,*) ''
  write(0,*) '********************** MAIN LOOP BEGINS! ***********************'
  write(0,*) ''
  loop_over_sources: do jsources = par%first_source,par%last_source
    !! Get number of geophone traces in shot gather (in some datasets is variable)
    par%ngeof = source_index(jsources,1)
    par%last_geof = par%first_geof+par%ngeof-1
  
    !! Create velocity frames for source wavefield propagation
    call vel%vshot(par,fwd%Vshot, par%sgx(:,jsources),.true.)
    call vel%vshot(par,fwd%Vshotr,par%sgx(:,jsources),.true.)
!    call sep_write(fwd%Vshot,'vrandom')
    if(.not. allocated(bwd%Vshot)) allocate(bwd%Vshot,mold=fwd%Vshot)
    bwd%Vshot=fwd%Vshot
    if(.not. allocated(bwd%Vshotr)) allocate(bwd%Vshotr,mold=fwd%Vshotr)
    bwd%Vshotr=fwd%Vshotr
  
    !! Create reflectivity frames for scattered wavefields
    call ref%rshot(par,wemva%pertshot,par%sgx(:,jsources))

    !! Compute shot position and report progress
    bwd%shot_pos = 1+floor((sx(par%trace_shot)-gx(par%trace_shot))/par%dx)
    write(0,'(a,i7,i7,i7,i7,i7,i7,i5)')     &
    'No. sources,curr source,shot pos.,source_index,first_geof,last_geof,ngeof',  &
    par%nsources,jsources,bwd%shot_pos,source_index(jsources,2),&
    par%first_geof,par%last_geof,par%ngeof
  
    !! Reading data (traces from one shot gather)
    call par%read_shot()

!    test: block
!      real, dimension(par%nkgx*par%nkz) :: data_, model
!  
!      if(adj) then
!        data_=1.0*reshape(wemva%pertshot,(/par%nkgx*par%nkz/))
!        model=0.
!      else
!        data_=0.
!        model=1.0*reshape(wemva%pertshot,(/par%nkgx*par%nkz/))
!      end if
!  
!      !! Initialize wemva function (2)
!      call wemva_op_init2(par%rcv,vel,ref,fwd,bwd,jsources)
  
      !! Perform WEMVA Modeling for each shot gather
      call wemva%wemva(par,fwd,bwd,vel,ref,jsources)
!      stat=wemva_op(adj,F,model,data_)
!  
!      if(adj) then
!        wemva%perturb=reshape(model,(/par%nkz,par%nkgx/))
!      else
!        wemva%perturb=reshape(data_,(/par%nkz,par%nkgx/))
!      end if
!
!      call wemva_op_deallocate()
!    end block test
  
    !! Construct the perturbation
    associate(x1=>1-par%napL,x2=>par%ngx+par%napR)
      call wemva%build_perturb(par,par%sgx(x1:x2,jsources))
    end associate

    !! Advance to the next receiver and the corresponding traces
    par%trace_shot = par%trace_shot+par%ngeof
    par%first_geof = par%last_geof+1
  end do loop_over_sources
  write(0,*) ''
  write(0,*) '********************** MAIN LOOP FINISHES! ***********************'
  write(0,*) ''
  
  !! Write the images history files and binaries
  call outp%write_history(par,image,wemva)
  
  !! Get the timing
  call DATE_AND_TIME(values=timer1)
  timer0 = timer1-timer0
  timer_sum = timer_sum + (((timer0(3)*24+timer0(5))*60+timer0(6))*60+timer0(7))*1000+timer0(8)
  call to_history("timer sum (sec)", timer_sum/1000)
  call to_history("timer sum (min)", timer_sum/1000/60)
  write(0,*) "timer sum (sec)", timer_sum/1000
  write(0,*) "timer sum (min)", timer_sum/1000/60
  
  !! Finale!
  write(0,*) ' '
  write(0,*) 'Finishing the wemva2d program normally!'

end program
